# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import List, Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from addict import Dict
from mmdet.structures.bbox import bbox_cxcywh_to_xyxy, bbox_xyxy_to_cxcywh
from mmengine.model import BaseModule
from torch import Tensor, nn

from mmtrack.registry import MODELS
from mmtrack.utils import OptConfigType, SampleList, max_last2d
from ..task_modules.filter import filter as filter_layer


@MODELS.register_module()
class PrDiMPClsHead(BaseModule):
    """PrDiMP classification head.

    Args:
        in_dim (int, optional): The dim of input feature. Defaults to 1024.
        out_dim (int, optional): The dim of output. Defaults to 512.
        filter_initializer (dict, optional): The configuration of filter
            initializer. Defaults to None.
        filter_optimizer (dict, optional): The configuration of filter
            optimizer. Defaults to None.
        locate_cfg (dict, optional): The configuration of bbox location.
            Defaults to None.
        update_cfg (dict, optional): The configuration of updating tracking
            state in memory. Defaults to None.
        optimizer_cfg (dict, optional): The configuration of optimizer.
            Defaults to None.
        loss_cls (dict, optional): The configuration of classification
            loss. Defaults to None.
        train_cfg (dict, optional): The configuration of training.
            Defaults to None.
    """

    def __init__(self,
                 in_dim: int = 1024,
                 out_dim: int = 512,
                 filter_initializer: OptConfigType = None,
                 filter_optimizer: OptConfigType = None,
                 locate_cfg: OptConfigType = None,
                 update_cfg: OptConfigType = None,
                 optimizer_cfg: OptConfigType = None,
                 loss_cls: OptConfigType = None,
                 train_cfg: OptConfigType = None,
                 test_cfg: OptConfigType = None,
                 **kwargs):
        super().__init__()
        filter_size = filter_initializer['filter_size']
        self.filter_initializer = MODELS.build(filter_initializer)
        self.filter_optimizer = MODELS.build(filter_optimizer)
        self.feat_norm_scale = math.sqrt(1.0 /
                                         (out_dim * filter_size * filter_size))
        self.channel_mapping = nn.Sequential(
            nn.Conv2d(in_dim, out_dim, kernel_size=3, padding=1, bias=False))

        self.locate_cfg = locate_cfg
        self.update_cfg = update_cfg
        self.optimizer_cfg = optimizer_cfg
        self.test_cfg = test_cfg
        self.train_cfg = train_cfg

        self.loss_cls = MODELS.build(loss_cls)

        if isinstance(filter_size, (int, float)):
            filter_size = [filter_size, filter_size]
        self.filter_size = torch.tensor(filter_size, dtype=torch.float32)
        self.feat_size = torch.tensor(
            train_cfg['feat_size'], dtype=torch.float32)
        self.img_size = torch.tensor(
            train_cfg['img_size'], dtype=torch.float32)

    def init_weights(self):
        """Initialize the parameters of this module."""
        for m in self.channel_mapping:
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
        self.filter_initializer.init_weights()

    def get_cls_feats(self, backbone_feats: Tensor) -> Tensor:
        """Get features for classification.

        Args:
            backbone_feats (Tensor): The features from backbone.

        Returns:
            Tensor: The features for classification.
        """
        cls_feats = self.channel_mapping(backbone_feats)
        scale_factor = (torch.tensor(cls_feats.shape[1:]).prod() / (torch.sum(
            (cls_feats**2).reshape(cls_feats.shape[0], 1, 1, -1),
            dim=3,
            keepdim=True) + 1e-5)).sqrt()
        cls_feats = cls_feats * self.feat_norm_scale * scale_factor

        return cls_feats

    def init_classifier(self,
                        backbone_feats: Tensor,
                        target_bboxes: Tensor,
                        dropout_probs: Optional[List] = None):
        """Initialize the filter and memory in the classifier.

        Args:
            backbone_feats (Tensor): The features from backbone.
            target_bboxes (Tensor): in [cx, cy, w, h] format.
            dropout_probs (list, optional): Defaults to None.
        """
        cls_feats = self.get_cls_feats(backbone_feats)

        # add features through the augmentation of `dropout`
        if dropout_probs is not None:
            aug_feats = []
            for i, prob in enumerate(dropout_probs):
                aug_feat = F.dropout2d(
                    cls_feats[:1, ...], p=prob, training=True)
                aug_feats.append(aug_feat)
            cls_feats = torch.cat([cls_feats] + aug_feats)

        # Get target filter by running the discriminative model prediction
        # module
        target_bboxes_xyxy = bbox_cxcywh_to_xyxy(target_bboxes)
        init_filter = self.filter_initializer(cls_feats, target_bboxes_xyxy)
        self.target_filter = self.filter_optimizer(
            init_filter,
            feat=cls_feats,
            bboxes=target_bboxes,
            num_iters=self.optimizer_cfg['init_update_iters'])

        # Initialize memory
        self.init_memory(cls_feats, target_bboxes)

    def init_memory(self, aug_feats: Tensor, target_bboxes: Tensor):
        """Initialize the some properties about training samples in memory:

            - `bboxes` (N, 4): the gt_bboxes of all samples in [cx, cy, w, h]
                format.
            - `training_samples` (N, C, H, W): the features of training samples
            - `sample_weights` (N,): the weights of all samples
            - `num_samples` (int): the number of all the samples fed into
                memory, including the outdated samples.
            - `replace_ind` (int): the index of samples in memory which would
                be replaced by the next new samples.

        Args:
            aug_feats (Tensor): The augmented features.
            target_bboxes (Tensor): of shape (N, 4) in [cx, cy, w, h] format.
        """
        self.memo = Dict()

        self.memo.bboxes = target_bboxes.new_zeros(
            self.update_cfg['sample_memory_size'], 4)
        self.memo.bboxes[:target_bboxes.shape[0], :] = target_bboxes

        # Initialize first-frame spatial training samples
        self.memo.num_samples = self.num_init_samples = aug_feats.size(0)
        # the index of the replaced memory samples in the next frame
        self.memo.replace_ind = None

        self.memo.sample_weights = aug_feats.new_zeros(
            self.update_cfg['sample_memory_size'])
        self.memo.sample_weights[:self.num_init_samples] = aug_feats.new_ones(
            1) / aug_feats.shape[0]

        self.memo.training_samples = aug_feats.new_zeros(
            self.update_cfg['sample_memory_size'], *aug_feats.shape[1:])
        self.memo.training_samples[:self.num_init_samples] = aug_feats

    def forward(self, backbone_feats: Tensor) -> Tuple[Tensor, Tensor]:
        """Run classifier on the backbone features.

        Args:
            backbone_feats (Tensor): the features from the last layer of
                backbone

        Returns:
            scores (Tensor): of shape (bs, 1, h, w)
            feats (Tensor): features for classification.
        """
        feats = self.get_cls_feats(backbone_feats)
        scores = filter_layer.apply_filter(feats, self.target_filter)
        return scores, feats

    def update_memory(self,
                      target_bbox: Tensor,
                      learning_rate: Optional[float] = None):
        """Update the tracking state in memory.

        Args:
            target_bbox (Tensor): of shape (1,4) in [x, y, w, h] format.
            learning_rate (float, optional): The learning rate about updating.
                Defaults to None.
        """
        # Update weights and get replace ind
        replace_ind = self.update_sample_weights(learning_rate)
        self.memo.replace_ind = replace_ind

        # Update training samples and bboxes in memory
        self.memo.training_samples[replace_ind:replace_ind + 1,
                                   ...] = self.memo.sample_feat
        self.memo.bboxes[replace_ind, :] = target_bbox
        self.memo.num_samples += 1

    def update_sample_weights(self, learning_rate=None) -> int:
        """Update the weights of samples in memory.

        Args:
            learning_rate (int, optional): The learning rate of updating
                samples in memory. Defaults to None.

        Returns:
            (int): the index of updated samples in memory.
        """

        init_sample_weight = self.update_cfg['init_samples_min_weight']
        if init_sample_weight == 0:
            init_sample_weight = None

        replace_start_ind = 0 if init_sample_weight is None else \
            self.num_init_samples

        if self.memo.num_samples == 0 or learning_rate == 1:
            self.memo.sample_weights[:] = 0
            self.memo.sample_weights[0] = 1
            replace_ind = 0
        else:
            # Get index to replace
            if self.memo.num_samples < self.memo.sample_weights.shape[0]:
                replace_ind = self.memo.num_samples
            else:
                _, replace_ind = torch.min(
                    self.memo.sample_weights[replace_start_ind:], 0)
                replace_ind = replace_ind.item() + replace_start_ind

            # Update weights
            if self.memo.replace_ind is None:
                self.memo.sample_weights /= 1 - learning_rate
                self.memo.sample_weights[replace_ind] = learning_rate
            else:
                self.memo.sample_weights[
                    replace_ind] = self.memo.sample_weights[
                        self.memo.replace_ind] / (1 - learning_rate)

        self.memo.sample_weights /= self.memo.sample_weights.sum()
        if (init_sample_weight is not None
                and self.memo.sample_weights[:self.num_init_samples].sum() <
                init_sample_weight):
            # TODO werid! the sum of samples_weights is not equal to 1.
            self.memo.sample_weights /= (
                init_sample_weight +
                self.memo.sample_weights[self.num_init_samples:].sum())
            self.memo.sample_weights[:self.num_init_samples] = (
                init_sample_weight / self.num_init_samples)

        return replace_ind

    def update_classifier(self,
                          target_bbox: Tensor,
                          frame_num: int,
                          hard_neg_flag: Optional[bool] = False):
        """Update the classifier with the refined bbox.

        Args:
            target_bbox (Tensor): of shape (1, 4) in [x, y, w, h] format.
            frame_num (int): The ID of frame.
            hard_neg_flag (bool, optional): Whether is the hard negative
                sample. Defaults to False.
        """
        # Set flags and learning rate
        learning_rate = self.update_cfg[
            'normal_lr'] if not hard_neg_flag else self.update_cfg[
                'hard_neg_lr']

        # Update the tracker memory
        if hard_neg_flag:
            self.update_memory(target_bbox, learning_rate)

        # Decide the number of iterations to run
        num_iters = 0
        if hard_neg_flag:
            num_iters = self.optimizer_cfg['hard_neg_iters']
        elif (frame_num - 1) % self.update_cfg['train_skipping'] == 0:
            num_iters = self.optimizer_cfg['update_iters']

        if num_iters > 0:
            # Get inputs for the DiMP filter optimizer module
            samples = self.memo.training_samples[:self.memo.num_samples, ...]
            target_bboxes = self.memo.bboxes[:self.memo.num_samples, :].clone()
            sample_weights = self.memo.sample_weights[:self.memo.num_samples]

            # Run the filter optimizer module
            self.target_filter = self.filter_optimizer(
                self.target_filter,
                num_iters=num_iters,
                feat=samples,
                bboxes=target_bboxes,
                sample_weights=sample_weights)

    def predict(self, backbone_feats: Tuple[Tensor], data_samples: SampleList,
                prev_bbox: Tensor, sample_center: Tensor,
                scale_factor: float) -> Tuple[Tensor, Tensor, bool]:
        """Perform forward propagation of the tracking head and predict
        tracking results on the features of the upstream network.

        Args:
            backbone_feats (Tuple[Tensor]): The features from backbone.
            data_samples (List[:obj:`TrackDataSample`]): The Data
                Samples. It usually includes information such as `gt_instance`.
            sample_center (Tensor): The coordinate of the sample center based
                on the original image.
            prev_bbox (Tensor): of shape (4, ) in [cx, cy, w, h] format.
            scale_factor (float): scale factor.

        Returns:
            new_bbox_center (Tensor): The center of the new bbox.
            scores_map (Tensor): The score map from the classifier.
            state (Tensor): The tracking state.
        """
        # ``self.memo.sample_feat`` is used to update the training samples
        # in the memory on some conditions.
        scores_raw, self.memo.sample_feat = self(backbone_feats[-1])
        scores_map = torch.softmax(
            scores_raw.view(-1), dim=0).view(scores_raw.shape)

        new_bbox_center, state = self.predict_by_feat(scores_map, prev_bbox,
                                                      sample_center,
                                                      scale_factor)

        return new_bbox_center, scores_map, state

    def _gen_2d_hanning_windows(self,
                                size: Sequence,
                                device: str = 'cuda') -> Tensor:
        """Generate 2D hanning window.

        Args:
            size (Sequence): The size of 2d hanning window.
            device (str): Device the tensor will be put on. Defaults to 'cuda'.

        Returns:
            Tensor: 2D hanning window with shape
            (num_base_anchors[i] * featmap_sizes[i][0] * featmap_sizes[i][1]).
        """

        def hanning_1d(s):
            return 0.5 * (1 - torch.cos(
                (2 * math.pi / (s + 1)) * torch.arange(1, s + 1).float()))

        hanning_win = hanning_1d(size[0]).reshape(-1, 1) * hanning_1d(
            size[1]).reshape(1, -1)

        return hanning_win.to(device)

    def predict_by_feat(self, scores: Tensor, prev_bbox: Tensor,
                        sample_center: Tensor,
                        scale_factor: float) -> Tuple[Tensor, bool]:
        """Track `prev_bbox` to current frame based on the output of network.

        Args:
            scores (Tensor): It's of shape (1, h, w) or (h, w).
            prev_bbox (Tensor): It's of shape (4,) in [cx, cy, w, h] format.
            sample_center (Tensor): The center of the cropped
                sample on the original image. It's of shape (1,2) or (2,) in
                [x, y] format.
            scale_factor (float): The scale of the cropped sample.
                It's of shape (1,) when it's a tensor.

        Return:
            Tensor: The displacement of the target to the center of original
                image
            bool: The tracking state.
        """
        sample_center = sample_center.squeeze()
        scores = scores.squeeze()
        prev_bbox = prev_bbox.squeeze()
        assert scores.dim() == 2
        assert sample_center.dim() == prev_bbox.dim() == 1

        score_size = torch.tensor([scores.shape[-1], scores.shape[-2]])
        output_size = (score_size - (self.filter_size + 1) % 2).to(
            scores.device)
        score_center = (score_size / 2).to(scores.device)

        scores_hn = scores
        if self.locate_cfg.get('hanning_window', False):
            scores_hn = scores * self._gen_2d_hanning_windows(
                score_size, device=scores.device)

        max_score, max_pos = max_last2d(scores_hn)
        max_pos = max_pos.flip(0).float()
        # the displacement of target to the center of score map
        target_disp_score_map = max_pos - score_center
        # the ratio of the size of original image to to that of score map
        ratio_size = (self.test_cfg['img_sample_size'] /
                      output_size) * scale_factor
        # the displcement of the target to the center of original image
        target_disp = target_disp_score_map * ratio_size

        # Handle different cases
        # 1. Target is not found
        if max_score.item() < self.locate_cfg['no_target_min_score']:
            return target_disp + sample_center, 'not_found'

        # Analysis whether there is a distractor
        # Calculate the size of neighborhood near the current target
        target_neigh_sz = self.locate_cfg['target_neighborhood_scale'] * (
            prev_bbox[2:4] / ratio_size)

        top_left = (max_pos - target_neigh_sz / 2).round().long()
        top_left = torch.clamp_min(top_left, 0).tolist()
        bottom_right = (max_pos + target_neigh_sz / 2).round().long()
        bottom_right = torch.clamp_max(bottom_right,
                                       score_size.min().item()).tolist()
        scores_masked = scores.clone()
        scores_masked[top_left[1]:bottom_right[1],
                      top_left[0]:bottom_right[0]] = 0

        # Find new maximum except the neighborhood of the target
        second_max_score, second_max_pos = max_last2d(scores_masked)
        second_max_pos = second_max_pos.flip(0).float().view(-1)
        distractor_disp_score_map = second_max_pos - score_center
        distractor_disp = distractor_disp_score_map * ratio_size
        # The displacement of previout target bbox to the center of the score
        # map.
        # Note that `sample_center`` may not be equal to the center of previous
        # tracking bbox due to different cropping mode
        prev_target_disp_score_map = (prev_bbox[:2] -
                                      sample_center) / ratio_size

        # 2. There is a distractor
        if second_max_score > self.locate_cfg['distractor_thres'] * max_score:
            target_disp_diff = torch.sqrt(
                torch.sum(
                    (target_disp_score_map - prev_target_disp_score_map)**2))
            # `distractor_disp_diff` is the displacement between current
            # tracking bbox and previous tracking bbox.
            distractor_disp_diff = torch.sqrt(
                torch.sum((distractor_disp_score_map -
                           prev_target_disp_score_map)**2))
            disp_diff_thres = self.locate_cfg[
                'dispalcement_scale'] * score_size.prod().float().sqrt() / 2

            if (distractor_disp_diff > disp_diff_thres
                    and target_disp_diff < disp_diff_thres):
                return target_disp + sample_center, 'hard_negative'
            if (distractor_disp_diff < disp_diff_thres
                    and target_disp_diff > disp_diff_thres):
                # The true target may be the `distractor` instead of the
                # `target` on this frame
                return distractor_disp + sample_center, 'hard_negative'
            else:
                # If both the displacement of target and distractor is larger
                # or smaller than the threshold, return the displacement of the
                # highest score.
                return target_disp + sample_center, 'uncertain'

        # 3. There is a hard negative object
        if (second_max_score > self.locate_cfg['hard_neg_thres'] * max_score
                and second_max_score > self.locate_cfg['no_target_min_score']):
            return target_disp + sample_center, 'hard_negative'

        # 4. Normal target
        return target_disp + sample_center, 'normal'

    def _gauss_1d(self,
                  size: Tensor,
                  sigma: float,
                  center: Tensor,
                  end_pad: int = 0,
                  return_density: bool = True):
        """Generate gauss labels.

        Args:
            size (Tensor): The size of score map with (2, ) shape.
            sigma (float): Standard deviations.
            center (Tensor): of (N, ) shape.
            end_pad (int, optional): The padding size.. Defaults to 0.
            return_density (bool, optional): Whether to return density.
                Defaults to True.

        Returns:
            Tensor: Gauss labels with (N, score_map_size) shape.
        """
        k = torch.arange(-(size - 1) / 2, (size + 1) / 2 + end_pad).reshape(
            1, -1).to(center.device)
        gauss = torch.exp(-1.0 / (2 * sigma**2) *
                          (k - center.reshape(-1, 1))**2)
        if not return_density:
            return gauss
        else:
            return gauss / (math.sqrt(2 * math.pi) * sigma)

    def _gauss_2d(self,
                  size: Tensor,
                  sigma: float,
                  center: Tensor,
                  end_pad: Union[Tensor, List, Tuple] = (0, 0),
                  return_density: bool = True):
        """Generate gauss labels.

        Args:
            size (Tensor): The size of score map with (2, ) shape.
            sigma (Tensor): Standard deviations.
            center (Tensor): The center of bbox with (N, 2) shape.
            end_pad (Union[Tensor, List, Tuple], optional): The padding size.
                Defaults to (0, 0).
            return_density (bool, optional): Whether to return density.
                Defaults to True.

        Returns:
            Tensor: Gauss labels with (N, score_map_size, score_map_size)
                shape.
        """
        if isinstance(sigma, (float, int)):
            sigma = (sigma, sigma)

        gauss_1d_out1 = self._gauss_1d(
            size[0].item(),
            sigma[0],
            center[:, 0],
            end_pad[0],
            return_density=return_density)
        gauss_1d_out1 = gauss_1d_out1.reshape(center.shape[0], 1, -1)

        gauss_1d_out2 = self._gauss_1d(
            size[1].item(),
            sigma[1],
            center[:, 1],
            end_pad[1],
            return_density=return_density)
        gauss_1d_out2 = gauss_1d_out2.reshape(center.shape[0], -1, 1)

        return gauss_1d_out1 * gauss_1d_out2

    def get_targets(self, target_center: Tensor) -> Tensor:
        """Generate the training targets for search images.

        Args:
            target_center (Tensor): The center of target bboxes with (N, 2)
                shape, in [x, y] format.

        Returns:
            Tensor: The distribution of gauss labels with
                (N, score_map_size, score_map_size) shape.
        """
        # TODO simplify the different devices
        img_size = self.img_size.to(target_center.device)
        filter_size = self.filter_size.to(target_center.device)
        feat_size = self.feat_size.to(target_center.device)

        # set the center of image as coordinate origin
        target_center_norm = (target_center - img_size / 2.) / img_size

        center = feat_size * target_center_norm + 0.5 * torch.fmod(
            filter_size + 1, 2)

        sigma = self.train_cfg['sigma_factor'] * feat_size.prod().sqrt().item()

        if self.train_cfg['end_pad_if_even']:
            end_pad = (torch.fmod(filter_size, 2) == 0).to(filter_size)
        else:
            end_pad = torch.zeros(2).to(filter_size)

        # generate gauss labels and density
        # Both of them are of shape (N, feat_size, feat_size)
        gauss_label_density = self._gauss_2d(
            feat_size, sigma, center, end_pad,
            self.train_cfg['use_gauss_density'])
        # continue to process label density
        feat_area = (feat_size + end_pad).prod()
        gauss_label_density = (1.0 - self.train_cfg['gauss_label_bias']
                               ) * gauss_label_density + self.train_cfg[
                                   'gauss_label_bias'] / feat_area

        mask = (gauss_label_density > self.train_cfg['label_density_threshold']
                ).to(gauss_label_density)
        gauss_label_density *= mask

        if self.train_cfg['label_density_norm']:
            g_sum = gauss_label_density.sum(dim=(-2, -1))
            valid = g_sum > 0.01
            gauss_label_density[valid, :, :] /= g_sum[valid].reshape(-1, 1, 1)
            gauss_label_density[~valid, :, :] = 1.0 / (
                gauss_label_density.shape[-2] * gauss_label_density.shape[-1])
        gauss_label_density *= 1.0 - self.train_cfg['label_density_shrink']

        return gauss_label_density

    def loss(self, template_feats: Tuple[Tensor], search_feats: Tuple[Tensor],
             batch_data_samples: SampleList, **kwargs) -> dict:
        """Perform forward propagation and loss calculation of the tracking
        head on the features of the upstream network.

        Args:
            template_feats (tuple[Tensor, ...]): Tuple of Tensor with
                shape (N, C, H, W) denoting the multi level feature maps of
                exemplar images.
            search_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of
                search images.
            batch_data_samples (List[:obj:`TrackDataSample`]): The Data
                Samples. It usually includes information such as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        batch_gt_bboxes = []
        batch_img_metas = []
        batch_search_gt_bboxes = []

        for data_sample in batch_data_samples:
            batch_img_metas.append(data_sample.metainfo)
            batch_gt_bboxes.append(data_sample.gt_instances['bboxes'])
            batch_search_gt_bboxes.append(
                data_sample.search_gt_instances['bboxes'])

        num_imgs_per_seq = batch_data_samples[0].gt_instances['bboxes'].shape[
            0]
        num_search_imgs_per_seq = batch_data_samples[0].search_gt_instances[
            'bboxes'].shape[0]

        # Extract features
        template_feats = self.get_cls_feats(template_feats[-1])
        template_feats = template_feats.reshape(num_imgs_per_seq, -1,
                                                *template_feats.shape[-3:])
        search_feats = self.get_cls_feats(search_feats[-1])
        search_feats = search_feats.reshape(num_search_imgs_per_seq, -1,
                                            *search_feats.shape[-3:])

        return self.loss_by_feat(template_feats, search_feats, batch_gt_bboxes,
                                 batch_search_gt_bboxes)

    def loss_by_feat(self, template_feats: Tensor, search_feats: Tensor,
                     batch_gt_bboxes: Tensor,
                     batch_search_gt_bboxes: Tensor) -> dict:
        """Compute loss.

        Args:
            modulations (Tuple[Tensor]): The modulation features.
            iou_feats (Tuple[Tensor]): The features for iou prediction.
            batch_gt_bboxes (Tensor): The gt_bboxes in a batch.
            batch_search_gt_bboxes (Tensor): The search gt_bboxes in a batch.

        Returns:
            dict: a dictionary of loss components.
        """
        # Train filter
        batch_gt_bboxes = torch.stack(batch_gt_bboxes, dim=1)
        # filter_iter is a list, and each of them is of shape
        # (bs, C, self.filter_size, self.filter_size)
        init_filter = self.filter_initializer(template_feats,
                                              batch_gt_bboxes.view(-1, 4))
        bboxes_cxcywh = bbox_xyxy_to_cxcywh(batch_gt_bboxes)
        _, filters_all_iters, _ = self.filter_optimizer(
            init_filter, feat=template_feats, bboxes=bboxes_cxcywh)

        # Classify samples using all filters
        # each item in ``target_scores`` is of shape
        # (num_search_imgs_per_seq, bs, score_map_size, score_map_size)
        target_scores = [
            filter_layer.apply_filter(search_feats, filter_weight)
            for filter_weight in filters_all_iters
        ]

        batch_search_gt_bboxes = torch.stack(batch_search_gt_bboxes, dim=1)
        search_gt_bboxes = batch_search_gt_bboxes.view(-1, 4)
        target_center = (search_gt_bboxes[:, :2] +
                         search_gt_bboxes[:, 2:4]) / 2.
        prob_labels_density = self.get_targets(target_center)
        # of shape (num_search_imgs_per_seq, bs, score_map_size, score_map_size) # noqa: E501
        prob_labels_density = prob_labels_density.view(
            batch_search_gt_bboxes.shape[0], -1,
            *prob_labels_density.shape[-2:])

        # compute loss
        loss_cls_list = [
            self.loss_cls(score, prob_labels_density, grid_dim=(-2, -1))
            for score in target_scores
        ]
        loss_cls_final = self.train_cfg['loss_weights'][
            'cls_final'] * loss_cls_list[-1]
        loss_cls_init = self.train_cfg['loss_weights'][
            'cls_init'] * loss_cls_list[0]

        if isinstance(self.train_cfg['loss_weights']['cls_iter'], list):
            loss_cls_iters = sum([
                weight * loss for weight, loss in zip(
                    self.train_cfg['loss_weights']['cls_iter'],
                    loss_cls_list[1:-1])
            ])
        else:
            loss_cls_iters = (self.train_cfg['loss_weights']['cls_iter'] /
                              (len(loss_cls_list) - 2)) * sum(
                                  loss_cls_list[1:-1])
        losses = dict(
            loss_cls_init=loss_cls_init,
            loss_cls_iter=loss_cls_iters,
            loss_cls_final=loss_cls_final)

        return losses



_base_ = [
    '../../_base_/models/faster-rcnn_r50_fpn.py',
    '../../_base_/datasets/mot_challenge.py', '../../_base_/default_runtime.py'
]
model = dict(
    type='DeepSORT',
    detector=dict(
        rpn_head=dict(bbox_coder=dict(clip_border=False)),
        roi_head=dict(
            bbox_head=dict(bbox_coder=dict(clip_border=False), num_classes=1)),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/faster_rcnn/faster-rcnn_r50_fpn_4e_mot17-half-64ee2ed4.pth'  # noqa: E501
        )),
    motion=dict(type='KalmanFilter', center_only=False),
    tracker=dict(
        type='SORTTracker', obj_score_thr=0.5, match_iou_thr=0.5, reid=None))

train_dataloader = None

train_cfg = None
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')



_base_ = [
    './strongsort_yolox_x_8xb4-80e_crowdhuman-mot17halftrain'
    '_test-mot17halfval.py'
]

model = dict(
    data_preprocessor=dict(
        type='TrackDataPreprocessor',
        pad_size_divisor=32,
        batch_augments=[
            dict(
                type='mmdet.BatchSyncRandomResize',
                random_size_range=(640, 1152))
        ]),
    detector=dict(
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/strongsort/mot_dataset/yolox_x_crowdhuman_mot20-private_20220812_192123-77c014de.pth'  # noqa: E501
        )),
    reid=dict(
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/reid/reid_r50_6e_mot20_20210803_212426-c83b1c01.pth'  # noqa: E501
        )),
)

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='mmdet.Resize', scale=(896, 1600), keep_ratio=True),
    dict(
        type='mmdet.Pad',
        size_divisor=32,
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='PackTrackInputs', pack_single_img=True)
]

val_dataloader = dict(
    dataset=dict(
        data_root='data/MOT17',
        ann_file='annotations/train_cocoformat.json',
        data_prefix=dict(img_path='train'),
        pipeline=test_pipeline))
test_dataloader = dict(
    dataset=dict(
        data_root='data/MOT20',
        ann_file='annotations/test_cocoformat.json',
        data_prefix=dict(img_path='test'),
        pipeline=test_pipeline))

test_evaluator = dict(format_only=True, outfile_prefix='./mot_20_test_res')



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import time
from functools import partial
from typing import List, Optional, Union

import torch
import torch.nn as nn
from mmcv.cnn import fuse_conv_bn
from mmengine import MMLogger
from mmengine.config import Config
from mmengine.device import get_max_cuda_memory
from mmengine.dist import get_world_size
from mmengine.runner import Runner, load_checkpoint
from mmengine.utils.dl_utils import set_multi_processing
from torch.nn.parallel import DistributedDataParallel

from mmtrack.datasets import BaseSOTDataset
from mmtrack.registry import DATASETS, MODELS

try:
    import psutil
except ImportError:
    psutil = None


def custom_round(value: Union[int, float],
                 factor: Union[int, float],
                 precision: int = 2) -> float:
    """Custom round function."""
    return round(value / factor, precision)


gb_round = partial(custom_round, factor=1024**3)


def print_log(msg: str, logger: Optional[MMLogger] = None) -> None:
    """Print a log message."""
    if logger is None:
        print(msg, flush=True)
    else:
        logger.info(msg)


def print_process_memory(p: 'psutil.Process',
                         logger: Optional[MMLogger] = None) -> None:
    """print process memory info."""
    mem_used = gb_round(psutil.virtual_memory().used)
    memory_full_info = p.memory_full_info()
    uss_mem = gb_round(memory_full_info.uss)
    pss_mem = gb_round(memory_full_info.pss)
    for children in p.children():
        child_mem_info = children.memory_full_info()
        uss_mem += gb_round(child_mem_info.uss)
        pss_mem += gb_round(child_mem_info.pss)
    process_count = 1 + len(p.children())
    print_log(
        f'(GB) mem_used: {mem_used:.2f} | uss: {uss_mem:.2f} | '
        f'pss: {pss_mem:.2f} | total_proc: {process_count}', logger)


class BaseBenchmark:
    """The benchmark base class.

    The ``run`` method is an external calling interface, and it will
    call the ``run_once`` method ``repeat_num`` times for benchmarking.
    Finally, call the ``average_multiple_runs`` method to further process
    the results of multiple runs.

    Args:
        max_iter (int): maximum iterations of benchmark.
        log_interval (int): interval of logging.
        num_warmup (int): Number of Warmup.
        logger (MMLogger, optional): Formatted logger used to record messages.
    """

    def __init__(self,
                 max_iter: int,
                 log_interval: int,
                 num_warmup: int,
                 logger: Optional[MMLogger] = None):
        self.max_iter = max_iter
        self.log_interval = log_interval
        self.num_warmup = num_warmup
        self.logger = logger

    def run(self, repeat_num: int = 1) -> dict:
        """benchmark entry method.

        Args:
            repeat_num (int): Number of repeat benchmark.
                Defaults to 1.
        """
        assert repeat_num >= 1

        results = []
        for _ in range(repeat_num):
            results.append(self.run_once())

        results = self.average_multiple_runs(results)
        return results

    def run_once(self) -> dict:
        """Executes the benchmark once."""
        raise NotImplementedError()

    def average_multiple_runs(self, results: List[dict]) -> dict:
        """Average the results of multiple runs."""
        raise NotImplementedError()


class InferenceBenchmark(BaseBenchmark):
    """The inference benchmark class. It will be statistical inference FPS,
    CUDA memory and CPU memory information.

    Args:
        cfg (mmengine.Config): config.
        checkpoint (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``.
        distributed (bool): distributed testing flag.
        is_fuse_conv_bn (bool): Whether to fuse conv and bn, this will
            slightly increase the inference speed.
        max_iter (int): maximum iterations of benchmark. Defaults to 2000.
        log_interval (int): interval of logging. Defaults to 50.
        num_warmup (int): Number of Warmup. Defaults to 5.
        logger (MMLogger, optional): Formatted logger used to record messages.
    """

    def __init__(self,
                 cfg: Config,
                 checkpoint: str,
                 distributed: bool,
                 is_fuse_conv_bn: bool,
                 max_iter: int = 2000,
                 log_interval: int = 50,
                 num_warmup: int = 5,
                 logger: Optional[MMLogger] = None):
        super().__init__(max_iter, log_interval, num_warmup, logger)

        assert get_world_size(
        ) == 1, 'Inference benchmark does not allow distributed multi-GPU'

        self.cfg = copy.deepcopy(cfg)
        self.distributed = distributed

        if psutil is None:
            raise ImportError('psutil is not installed, please install it by: '
                              'pip install psutil')

        self._process = psutil.Process()
        env_cfg = self.cfg.get('env_cfg')
        if env_cfg.get('cudnn_benchmark'):
            torch.backends.cudnn.benchmark = True

        mp_cfg: dict = env_cfg.get('mp_cfg', {})
        set_multi_processing(**mp_cfg, distributed=self.distributed)

        print_log('before build: ', self.logger)
        print_process_memory(self._process, self.logger)

        self.model = self._init_model(checkpoint, is_fuse_conv_bn)

        # Because multiple processes will occupy additional CPU resources,
        # FPS statistics will be more unstable when num_workers is not 0.
        # It is reasonable to set num_workers to 0.
        dataloader_cfg = cfg.test_dataloader
        dataloader_cfg['num_workers'] = 0
        dataloader_cfg['batch_size'] = 1
        dataloader_cfg['persistent_workers'] = False
        self.data_loader = Runner.build_dataloader(dataloader_cfg)

        print_log('after build: ', self.logger)
        print_process_memory(self._process, self.logger)

    def _init_model(self, checkpoint: str, is_fuse_conv_bn: bool) -> nn.Module:
        """Initialize the model."""
        model = MODELS.build(self.cfg.model)
        if checkpoint is not None:
            load_checkpoint(model, checkpoint, map_location='cpu')
        if is_fuse_conv_bn:
            model = fuse_conv_bn(model)

        model = model.cuda()

        if self.distributed:
            model = DistributedDataParallel(
                model,
                device_ids=[torch.cuda.current_device()],
                broadcast_buffers=False,
                find_unused_parameters=False)

        model.eval()
        return model

    def run_once(self) -> dict:
        """Executes the benchmark once."""
        pure_inf_time = 0
        fps = 0

        for i, data in enumerate(self.data_loader):

            if (i + 1) % self.log_interval == 0:
                print_log('==================================', self.logger)

            torch.cuda.synchronize()
            start_time = time.perf_counter()

            with torch.no_grad():
                self.model.test_step(data)

            torch.cuda.synchronize()
            elapsed = time.perf_counter() - start_time

            if i >= self.num_warmup:
                pure_inf_time += elapsed
                if (i + 1) % self.log_interval == 0:
                    fps = (i + 1 - self.num_warmup) / pure_inf_time
                    cuda_memory = get_max_cuda_memory()

                    print_log(
                        f'Done image [{i + 1:<3}/{self.max_iter}], '
                        f'fps: {fps:.1f} img/s, '
                        f'times per image: {1000 / fps:.1f} ms/img, '
                        f'cuda memory: {cuda_memory} MB', self.logger)
                    print_process_memory(self._process, self.logger)

            if (i + 1) == self.max_iter:
                fps = (i + 1 - self.num_warmup) / pure_inf_time
                break

        return {'fps': fps}

    def average_multiple_runs(self, results: List[dict]) -> dict:
        """Average the results of multiple runs."""
        print_log('============== Done ==================', self.logger)

        fps_list_ = [round(result['fps'], 1) for result in results]
        avg_fps_ = sum(fps_list_) / len(fps_list_)
        outputs = {'avg_fps': avg_fps_, 'fps_list': fps_list_}

        if len(fps_list_) > 1:
            times_pre_image_list_ = [
                round(1000 / result['fps'], 1) for result in results
            ]
            avg_times_pre_image_ = sum(times_pre_image_list_) / len(
                times_pre_image_list_)

            print_log(
                f'Overall fps: {fps_list_}[{avg_fps_:.1f}] img/s, '
                'times per image: '
                f'{times_pre_image_list_}[{avg_times_pre_image_:.1f}] '
                'ms/img', self.logger)
        else:
            print_log(
                f'Overall fps: {fps_list_[0]:.1f} img/s, '
                f'times per image: {1000 / fps_list_[0]:.1f} ms/img',
                self.logger)

        print_log(f'cuda memory: {get_max_cuda_memory()} MB', self.logger)
        print_process_memory(self._process, self.logger)

        return outputs


class DataLoaderBenchmark(BaseBenchmark):
    """The dataloader benchmark class. It will be statistical inference FPS and
    CPU memory information.

    Args:
        cfg (mmengine.Config): config.
        distributed (bool): distributed testing flag.
        dataset_type (str): benchmark data type, only supports ``train``,
            ``val`` and ``test``.
        max_iter (int): maximum iterations of benchmark. Defaults to 2000.
        log_interval (int): interval of logging. Defaults to 50.
        num_warmup (int): Number of Warmup. Defaults to 5.
        logger (MMLogger, optional): Formatted logger used to record messages.
    """

    def __init__(self,
                 cfg: Config,
                 distributed: bool,
                 dataset_type: str,
                 max_iter: int = 2000,
                 log_interval: int = 50,
                 num_warmup: int = 5,
                 logger: Optional[MMLogger] = None):
        super().__init__(max_iter, log_interval, num_warmup, logger)

        assert dataset_type in ['train', 'val', 'test'], \
            'dataset_type only supports train,' \
            f' val and test, but got {dataset_type}'
        assert get_world_size(
        ) == 1, 'Dataloader benchmark does not allow distributed multi-GPU'

        self.cfg = copy.deepcopy(cfg)
        self.distributed = distributed

        if psutil is None:
            raise ImportError('psutil is not installed, please install it by: '
                              'pip install psutil')
        self._process = psutil.Process()

        mp_cfg = self.cfg.get('env_cfg', {}).get('mp_cfg')
        if mp_cfg is not None:
            set_multi_processing(distributed=self.distributed, **mp_cfg)
        else:
            set_multi_processing(distributed=self.distributed)

        print_log('before build: ', self.logger)
        print_process_memory(self._process, self.logger)

        if dataset_type == 'train':
            self.data_loader = Runner.build_dataloader(cfg.train_dataloader)
        elif dataset_type == 'test':
            self.data_loader = Runner.build_dataloader(cfg.test_dataloader)
        else:
            self.data_loader = Runner.build_dataloader(cfg.val_dataloader)

        self.batch_size = self.data_loader.batch_size
        self.num_workers = self.data_loader.num_workers

        print_log('after build: ', self.logger)
        print_process_memory(self._process, self.logger)

    def run_once(self) -> dict:
        """Executes the benchmark once."""
        pure_inf_time = 0
        fps = 0

        # benchmark with 2000 image and take the average
        start_time = time.perf_counter()
        for i, data in enumerate(self.data_loader):
            elapsed = time.perf_counter() - start_time

            if (i + 1) % self.log_interval == 0:
                print_log('==================================', self.logger)

            if i >= self.num_warmup:
                pure_inf_time += elapsed
                if (i + 1) % self.log_interval == 0:
                    fps = (i + 1 - self.num_warmup) / pure_inf_time

                    print_log(
                        f'Done batch [{i + 1:<3}/{self.max_iter}], '
                        f'fps: {fps:.1f} batch/s, '
                        f'times per batch: {1000 / fps:.1f} ms/batch, '
                        f'batch size: {self.batch_size}, num_workers: '
                        f'{self.num_workers}', self.logger)
                    print_process_memory(self._process, self.logger)

            if (i + 1) == self.max_iter:
                fps = (i + 1 - self.num_warmup) / pure_inf_time
                break

            start_time = time.perf_counter()

        return {'fps': fps}

    def average_multiple_runs(self, results: List[dict]) -> dict:
        """Average the results of multiple runs."""
        print_log('============== Done ==================', self.logger)

        fps_list_ = [round(result['fps'], 1) for result in results]
        avg_fps_ = sum(fps_list_) / len(fps_list_)
        outputs = {'avg_fps': avg_fps_, 'fps_list': fps_list_}

        if len(fps_list_) > 1:
            times_pre_image_list_ = [
                round(1000 / result['fps'], 1) for result in results
            ]
            avg_times_pre_image_ = sum(times_pre_image_list_) / len(
                times_pre_image_list_)

            print_log(
                f'Overall fps: {fps_list_}[{avg_fps_:.1f}] img/s, '
                'times per batch: '
                f'{times_pre_image_list_}[{avg_times_pre_image_:.1f}] '
                f'ms/batch, batch size: {self.batch_size}, num_workers: '
                f'{self.num_workers}', self.logger)
        else:
            print_log(
                f'Overall fps: {fps_list_[0]:.1f} batch/s, '
                f'times per batch: {1000 / fps_list_[0]:.1f} ms/batch, '
                f'batch size: {self.batch_size}, num_workers: '
                f'{self.num_workers}', self.logger)

        print_process_memory(self._process, self.logger)

        return outputs


class DatasetBenchmark(BaseBenchmark):
    """The dataset benchmark class. It will be statistical inference FPS, FPS
    pre transform and CPU memory information.

    Args:
        cfg (mmengine.Config): config.
        dataset_type (str): benchmark data type, only supports ``train``,
            ``val`` and ``test``.
        max_iter (int): maximum iterations of benchmark. Defaults to 2000.
        log_interval (int): interval of logging. Defaults to 50.
        num_warmup (int): Number of Warmup. Defaults to 5.
        logger (MMLogger, optional): Formatted logger used to record messages.
    """

    def __init__(self,
                 cfg: Config,
                 dataset_type: str,
                 max_iter: int = 2000,
                 log_interval: int = 50,
                 num_warmup: int = 5,
                 logger: Optional[MMLogger] = None):
        super().__init__(max_iter, log_interval, num_warmup, logger)
        assert dataset_type in ['train', 'val', 'test'], \
            'dataset_type only supports train,' \
            f' val and test, but got {dataset_type}'
        assert get_world_size(
        ) == 1, 'Dataset benchmark does not allow distributed multi-GPU'
        self.cfg = copy.deepcopy(cfg)

        if dataset_type == 'train':
            dataloader_cfg = copy.deepcopy(cfg.train_dataloader)
        elif dataset_type == 'test':
            dataloader_cfg = copy.deepcopy(cfg.test_dataloader)
        else:
            dataloader_cfg = copy.deepcopy(cfg.val_dataloader)

        dataset_cfg = dataloader_cfg.pop('dataset')
        dataset = DATASETS.build(dataset_cfg)
        if hasattr(dataset, 'full_init'):
            dataset.full_init()
        self.dataset = dataset
        self.dataset_type = dataset_type

    def run_once(self) -> dict:
        """Executes the benchmark once."""
        pure_inf_time = 0
        fps = 0

        if self.dataset_type == 'test' and isinstance(self.dataset,
                                                      BaseSOTDataset):
            total_index = []
            for video_ind in range(self.dataset.num_videos):
                total_index.extend([
                    (video_ind, frame_ind) for frame_ind in range(
                        self.dataset.get_len_per_video(video_ind))
                ])
        else:
            total_index = list(range(len(self.dataset)))

        start_time = time.perf_counter()
        for i, idx in enumerate(total_index):
            if (i + 1) % self.log_interval == 0:
                print_log('==================================', self.logger)

            get_data_info_start_time = time.perf_counter()
            valid_idx = idx[0] if isinstance(idx, tuple) else idx
            self.dataset.get_data_info(valid_idx)
            get_data_info_elapsed = time.perf_counter(
            ) - get_data_info_start_time

            if (i + 1) % self.log_interval == 0:
                print_log(f'get_data_info - {get_data_info_elapsed * 1000} ms',
                          self.logger)

            self.dataset[idx]
            elapsed = time.perf_counter() - start_time - get_data_info_elapsed

            if i >= self.num_warmup:
                # print_log(f'{elapsed}', self.logger)
                pure_inf_time += elapsed
                if (i + 1) % self.log_interval == 0:
                    fps = (i + 1 - self.num_warmup) / pure_inf_time

                    print_log(
                        f'Done img [{i + 1:<3}/{self.max_iter}], '
                        f'fps: {fps:.1f} img/s, '
                        f'times per img: {1000 / fps:.1f} ms/img', self.logger)

            if (i + 1) == self.max_iter:
                fps = (i + 1 - self.num_warmup) / pure_inf_time
                break

            start_time = time.perf_counter()

        return {'fps': fps}

    def average_multiple_runs(self, results: List[dict]) -> dict:
        """Average the results of multiple runs."""
        print_log('============== Done ==================', self.logger)

        fps_list_ = [round(result['fps'], 1) for result in results]
        avg_fps_ = sum(fps_list_) / len(fps_list_)
        outputs = {'avg_fps': avg_fps_, 'fps_list': fps_list_}

        if len(fps_list_) > 1:
            times_pre_image_list_ = [
                round(1000 / result['fps'], 1) for result in results
            ]
            avg_times_pre_image_ = sum(times_pre_image_list_) / len(
                times_pre_image_list_)

            print_log(
                f'Overall fps: {fps_list_}[{avg_fps_:.1f}] img/s, '
                'times per img: '
                f'{times_pre_image_list_}[{avg_times_pre_image_:.1f}] '
                'ms/img', self.logger)
        else:
            print_log(
                f'Overall fps: {fps_list_[0]:.1f} img/s, '
                f'times per img: {1000 / fps_list_[0]:.1f} ms/img',
                self.logger)

        return outputs



_base_ = [
    '../../_base_/models/mask-rcnn_r50_fpn.py',
    '../../_base_/datasets/youtube_vis.py', '../../_base_/default_runtime.py'
]
model = dict(
    type='MaskTrackRCNN',
    detector=dict(
        roi_head=dict(
            bbox_head=dict(num_classes=40), mask_head=dict(num_classes=40)),
        train_cfg=dict(
            rpn=dict(sampler=dict(num=64)),
            rpn_proposal=dict(nms_pre=200, max_per_img=200),
            rcnn=dict(sampler=dict(num=128))),
        test_cfg=dict(
            rpn=dict(nms_pre=200, max_per_img=200), rcnn=dict(score_thr=0.01)),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmdetection/v2.0/mask_rcnn/mask_rcnn_r50_fpn_1x_coco/mask_rcnn_r50_fpn_1x_coco_20200205-d4b0c5d6.pth'  # noqa: E501
        )),
    track_head=dict(
        type='RoITrackHead',
        roi_extractor=dict(
            type='SingleRoIExtractor',
            _scope_='mmdet',
            roi_layer=dict(type='RoIAlign', output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32]),
        embed_head=dict(
            type='RoIEmbedHead',
            num_fcs=2,
            roi_feat_size=7,
            in_channels=256,
            fc_out_channels=1024),
        train_cfg=dict(
            assigner=dict(
                type='MaxIoUAssigner',
                _scope_='mmdet',
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=True,
                ignore_iof_thr=-1),
            sampler=dict(
                type='RandomSampler',
                _scope_='mmdet',
                num=128,
                pos_fraction=0.25,
                neg_pos_ub=-1,
                add_gt_as_proposals=True),
            pos_weight=-1,
            debug=False)),
    tracker=dict(
        type='MaskTrackRCNNTracker',
        match_weights=dict(det_score=1.0, iou=2.0, det_label=10.0),
        num_frames_retain=20))

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.00125, momentum=0.9, weight_decay=0.0001),
    clip_grad=dict(max_norm=35, norm_type=2))

# learning policy
param_scheduler = [
    dict(
        type='mmdet.LinearLR',
        start_factor=1.0 / 3.0,
        by_epoch=False,
        begin=0,
        end=500),
    dict(
        type='mmdet.MultiStepLR',
        begin=0,
        end=12,
        by_epoch=True,
        milestones=[8, 11],
        gamma=0.1)
]
# runtime settings
train_cfg = dict(type='EpochBasedTrainLoop', max_epochs=12, val_begin=13)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# evaluator
val_evaluator = dict(
    type='YouTubeVISMetric',
    metric='youtube_vis_ap',
    outfile_prefix='./youtube_vis_results',
    format_only=True)
test_evaluator = val_evaluator



_base_ = [
    '../../_base_/datasets/vot2018.py',
    './siamese-rpn_r50_8xb28-20e_imagenetvid-imagenetdet-coco_base.py'
]

# model settings
model = dict(
    test_cfg=dict(
        rpn=dict(penalty_k=0.04, window_influence=0.44, lr=0.33),
        test_mode='VOT'))



# Copyright (c) OpenMMLab. All rights reserved.
# The codes are modified from https://github.com/visionml/pytracking/blob/master/ltr/models/layers/filter.py # noqa: E501
import torch.nn.functional as F
from torch import Tensor


def apply_filter(feat: Tensor, filter_weights: Tensor) -> Tensor:
    """Applies the filter on the input features.

    The number of groups is automatically calculated.
    args:
        feat (Tensor): The input features with two possible shapes in the
            different modes:
            - training mode: of shape (num_img_per_seq, bs, c, h, w)
            - test mode: of shape (num_img_per_seq, c, h, w).
        filter_weights (Tensor): The filter to be applied on the `feat`. There
            are two possible shapes in the different modes:
            - training mode: of shape (bs, c, filter_h, filter_w)
            - test mode: of shape (1, c, filter_h, filter_w)
    output:
        scores (Tenosr): Output of filtering.
            - train mode: (num_img_per_seq, bs, h, w)
            - test mode: (num_img_per_seq, 1, h, w)
    """
    padding = (filter_weights.shape[-2] // 2, filter_weights.shape[-1] // 2)
    num_groups = feat.shape[1] if feat.dim() == 5 else 1
    scores = F.conv2d(
        feat.reshape(feat.shape[0], -1, *feat.shape[-2:]),
        filter_weights,
        padding=padding,
        groups=num_groups)
    return scores


def apply_feat_transpose(feat: Tensor,
                         activation: Tensor,
                         filter_size_hw: Tensor,
                         training: bool = True) -> Tensor:
    """The transposed operation of `apply_filter` w.r.t the filter. It can be
    used to compute the filter gradient. There are two implements: the one
    forwards slowly and backwards fast, which used in training, and the other
    is the opposite, which used in test.

    Args:
        feat (Tensor): The input features with two possible shapes in the
            different modes:
            - training mode: of shape (num_img_per_seq, bs, c, h, w)
            - test mode: of shape (num_img_per_seq, c, h, w).
        activation (Tensor): The activation (e.g. residuals between output and
            label). There are two possible shapes in the different modes:
            - training mode: of shape (num_img_per_seq, bs, size_h, size_w)
            - test mode: of shape (num_img_per_seq, 1, size_h, size_w)
        filter_size_hw (Tensor): of shape (2,) shape in [h, w] format.
        training (bool, optional): Whether training mode or not. The faster
            implementation is chose according to this.

    Returns:
        Tensor. There are two possible shape in the
            different mode:
            - training mode: of shape (bs, c, filter_h, fiter_w).
            - test mode: of shape (1, c, filter_h, fiter_w).
    """

    if isinstance(filter_size_hw, int):
        filter_size_hw = (filter_size_hw, filter_size_hw)

    if training:
        # slow forward and fast backward
        # TODO: check the pad difference in training and test mode
        transpose_pad = [sz // 2 for sz in filter_size_hw]
        num_img_per_seq = feat.shape[0]
        batch_size = feat.shape[1] if feat.dim() == 5 else 1

        filter_grad = F.conv2d(
            feat.reshape(-1, *feat.shape[-3:]).permute(1, 0, 2, 3),
            activation.reshape(-1, 1, *activation.shape[-2:]),
            padding=transpose_pad,
            groups=num_img_per_seq * batch_size)

        # The shape of returns is (bs, c, filter_h, fiter_w)
        if num_img_per_seq == 1:
            return filter_grad.permute(1, 0, 2, 3)

        filter_grad = filter_grad.view(-1, num_img_per_seq, batch_size,
                                       *filter_grad.shape[-2:])
        return filter_grad.sum(dim=1).permute(1, 0, 2, 3)
    else:
        # fast forwward and slow backward
        transpose_pad = [(sz - 1) // 2 for sz in filter_size_hw]
        batch_size = feat.shape[0]
        filter_grad = F.conv2d(
            activation.reshape(1, -1, *activation.shape[-2:]),
            feat.reshape(-1, 1, *feat.shape[-2:]),
            padding=transpose_pad,
            groups=batch_size)

        filter_grad = filter_grad.view(batch_size, 1, -1,
                                       *filter_grad.shape[-2:])
        # The shape of returns is (1, c, filter_h, fiter_w)
        return filter_grad.sum(dim=0).flip((2, 3))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Union

from torch import Tensor

from mmtrack.models.mot import BaseMultiObjectTracker
from mmtrack.registry import MODELS
from mmtrack.utils import OptConfigType, OptMultiConfig, SampleList


@MODELS.register_module()
class Mask2Former(BaseMultiObjectTracker):
    r"""Implementation of `Masked-attention Mask
    Transformer for Universal Image Segmentation
    <https://arxiv.org/pdf/2112.01527>`_.

    Args:
        backbone (dict): Configuration of backbone. Defaults to None.
        track_head (dict): Configuration of track head. Defaults to None.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
            Defaults to None.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 backbone: Optional[dict] = None,
                 track_head: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super(BaseMultiObjectTracker, self).__init__(
            data_preprocessor=data_preprocessor, init_cfg=init_cfg)

        if backbone is not None:
            self.backbone = MODELS.build(backbone)

        if track_head is not None:
            self.track_head = MODELS.build(track_head)

        self.num_classes = self.track_head.num_classes

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        """Overload in order to load mmdet pretrained ckpt."""
        for key in list(state_dict):
            if key.startswith('panoptic_head'):
                state_dict[key.replace('panoptic',
                                       'track')] = state_dict.pop(key)

        super()._load_from_state_dict(state_dict, prefix, local_metadata,
                                      strict, missing_keys, unexpected_keys,
                                      error_msgs)

    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> Union[dict, tuple]:
        """
        Args:
            inputs (Tensor): Input images of shape (N, T, C, H, W).
                These should usually be mean centered and std scaled.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        # shape (N * T, C, H, W)
        img = img.flatten(0, 1)

        x = self.backbone(img)
        losses = self.track_head.loss(x, data_samples)

        return losses

    def predict(self,
                inputs: dict,
                data_samples: SampleList,
                rescale: bool = True) -> SampleList:
        """Predict results from a batch of inputs and data samples with
        postprocessing.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W)
                encoding input images. Typically, these should be mean centered
                and std scaled. The N denotes batch size. The T denotes the
                number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
                In test mode, T = 1 and there is only ``img`` and no
                ``ref_img``.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            SampleList: Tracking results of the input images.
            Each TrackDataSample usually contains ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        # the "T" is 1
        img = img.squeeze(1)
        feats = self.backbone(img)
        pred_track_ins_list = self.track_head.predict(feats, data_samples,
                                                      rescale)

        results = []
        for idx, pred_track_ins in enumerate(pred_track_ins_list):
            track_data_sample = data_samples[idx]
            track_data_sample.pred_track_instances = pred_track_ins
            results.append(track_data_sample)

        return results



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
from mmcls.evaluation.metrics import Accuracy
from mmengine.model import BaseModule

from mmtrack.registry import MODELS
from mmtrack.structures import ReIDDataSample
from .fc_module import FcModule


@MODELS.register_module()
class LinearReIDHead(BaseModule):
    """Linear head for re-identification.

    Args:
        num_fcs (int): Number of fcs.
        in_channels (int): Number of channels in the input.
        fc_channels (int): Number of channels in the fcs.
        out_channels (int): Number of channels in the output.
        norm_cfg (dict, optional): Configuration of normlization method
            after fc. Defaults to None.
        act_cfg (dict, optional): Configuration of activation method after fc.
            Defaults to None.
        num_classes (int, optional): Number of the identities. Default to None.
        loss_cls (dict, optional): Cross entropy loss to train the ReID module.
            Defaults to None.
        loss_triplet (dict, optional): Triplet loss to train the ReID module.
            Defaults to None.
        topk (int | Tuple[int]): Top-k accuracy. Defaults to ``(1, )``.
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Defaults to dict(type='Normal',layer='Linear', mean=0, std=0.01,
            bias=0).
    """

    def __init__(self,
                 num_fcs: int,
                 in_channels: int,
                 fc_channels: int,
                 out_channels: int,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: Optional[dict] = None,
                 num_classes: Optional[int] = None,
                 loss_cls: Optional[dict] = None,
                 loss_triplet: Optional[dict] = None,
                 topk: Union[int, Tuple[int]] = (1, ),
                 init_cfg: Union[dict, List[dict]] = dict(
                     type='Normal', layer='Linear', mean=0, std=0.01, bias=0)):
        super(LinearReIDHead, self).__init__(init_cfg=init_cfg)

        assert isinstance(topk, (int, tuple))
        if isinstance(topk, int):
            topk = (topk, )
        for _topk in topk:
            assert _topk > 0, 'Top-k should be larger than 0'
        self.topk = topk

        if loss_cls is None:
            if isinstance(num_classes, int):
                warnings.warn('Since cross entropy is not set, '
                              'the num_classes will be ignored.')
            if loss_triplet is None:
                raise ValueError('Please choose at least one loss in '
                                 'triplet loss and cross entropy loss.')
        elif not isinstance(num_classes, int):
            raise TypeError('The num_classes must be a current number, '
                            'if there is cross entropy loss.')
        self.loss_cls = MODELS.build(loss_cls) if loss_cls else None
        self.loss_triplet = MODELS.build(loss_triplet) \
            if loss_triplet else None

        self.num_fcs = num_fcs
        self.in_channels = in_channels
        self.fc_channels = fc_channels
        self.out_channels = out_channels
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.num_classes = num_classes

        self._init_layers()

    def _init_layers(self):
        """Initialize fc layers."""
        self.fcs = nn.ModuleList()
        for i in range(self.num_fcs):
            in_channels = self.in_channels if i == 0 else self.fc_channels
            self.fcs.append(
                FcModule(in_channels, self.fc_channels, self.norm_cfg,
                         self.act_cfg))
        in_channels = self.in_channels if self.num_fcs == 0 else \
            self.fc_channels
        self.fc_out = nn.Linear(in_channels, self.out_channels)
        if self.loss_cls:
            self.bn = nn.BatchNorm1d(self.out_channels)
            self.classifier = nn.Linear(self.out_channels, self.num_classes)

    def forward(self, feats: Tuple[torch.Tensor]) -> torch.Tensor:
        """The forward process."""
        # Multiple stage inputs are acceptable
        # but only the last stage will be used.
        feats = feats[-1]

        for m in self.fcs:
            feats = m(feats)
        feats = self.fc_out(feats)
        return feats

    def loss(self, feats: Tuple[torch.Tensor],
             data_samples: List[ReIDDataSample]) -> dict:
        """Calculate losses.

        Args:
            feats (tuple[Tensor]): The features extracted from the backbone.
            data_samples (List[ReIDDataSample]): The annotation data of
                every samples.

        Returns:
            dict: a dictionary of loss components
        """
        # The part can be traced by torch.fx
        feats = self(feats)

        # The part can not be traced by torch.fx
        losses = self.loss_by_feat(feats, data_samples)
        return losses

    def loss_by_feat(self, feats: torch.Tensor,
                     data_samples: List[ReIDDataSample]) -> dict:
        """Unpack data samples and compute loss."""
        losses = dict()
        gt_label = torch.cat([i.gt_label.label for i in data_samples])

        if self.loss_triplet:
            losses['triplet_loss'] = self.loss_triplet(feats, gt_label)

        if self.loss_cls:
            feats_bn = self.bn(feats)
            cls_score = self.classifier(feats_bn)
            losses['ce_loss'] = self.loss_cls(cls_score, gt_label)
            acc = Accuracy.calculate(cls_score, gt_label, topk=self.topk)
            losses.update(
                {f'accuracy_top-{k}': a
                 for k, a in zip(self.topk, acc)})

        return losses

    def predict(
            self,
            feats: Tuple[torch.Tensor],
            data_samples: List[ReIDDataSample] = None) -> List[ReIDDataSample]:
        """Inference without augmentation.

        Args:
            feats (Tuple[Tensor]): The features extracted from the backbone.
                Multiple stage inputs are acceptable but only the last stage
                will be used.
            data_samples (List[ReIDDataSample], optional): The annotation
                data of every samples. If not None, set ``pred_label`` of
                the input data samples. Defaults to None.

        Returns:
            List[ReIDDataSample]: A list of data samples which contains the
            predicted results.
        """
        # The part can be traced by torch.fx
        feats = self(feats)

        # The part can not be traced by torch.fx
        data_samples = self.predict_by_feat(feats, data_samples)

        return data_samples

    def predict_by_feat(
            self,
            feats: torch.Tensor,
            data_samples: List[ReIDDataSample] = None) -> List[ReIDDataSample]:
        """Add prediction features to data samples."""
        if data_samples is not None:
            for data_sample, feat in zip(data_samples, feats):
                data_sample.pred_feature = feat
        else:
            data_samples = []
            for feat in feats:
                data_sample = ReIDDataSample()
                data_sample.pred_feature = feat
                data_samples.append(data_sample)

        return data_samples



# Copyright (c) OpenMMLab. All rights reserved.
import os
import time
from typing import List

from mmtrack.registry import DATASETS
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class UAV123Dataset(BaseSOTDataset):
    """UAV123 dataset of single object tracking.

    The dataset is only used to test.
    """

    def __init__(self, *args, **kwargs):
        """Initialization of SOT dataset class."""
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load dataset information.

        Returns:
            list[dict]: The length of the list is the number of videos. The
                inner dict is in the following format:
                    {
                        'video_path': the video path
                        'ann_path': the annotation path
                        'start_frame_id': the starting frame number contained
                            in the image name
                        'end_frame_id': the ending frame number contained in
                            the image name
                        'framename_template': the template of image name
                    }
        """
        print('Loading UAV123 dataset...')
        start_time = time.time()
        data_infos = []
        data_infos_str = self._loadtxt(
            self.ann_file, return_ndarray=False).split('\n')
        # the first line of annotation file is a dataset comment.
        for line in data_infos_str[1:]:
            # compatible with different OS.
            line = line.strip().replace('/', os.sep).split(',')
            data_info = dict(
                video_path=line[0],
                ann_path=line[1],
                start_frame_id=int(line[2]),
                end_frame_id=int(line[3]),
                framename_template='%06d.jpg')
            data_infos.append(data_info)
        print(f'UAV123 dataset loaded! ({time.time()-start_time:.2f} s)')
        return data_infos



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import List, Optional, Tuple, Union

import torch
from addict import Dict
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.utils import (ConfigType, OptConfigType, SampleList,
                           convert_data_sample_type)
from .base import BaseVideoDetector


@MODELS.register_module()
class SELSA(BaseVideoDetector):
    """Sequence Level Semantics Aggregation for Video Object Detection.

    This video object detector is the implementation of `SELSA
    <https://arxiv.org/abs/1907.06390>`_.
    """

    def __init__(self,
                 detector: ConfigType,
                 frozen_modules: Optional[Union[List[str], Tuple[str],
                                                str]] = None,
                 train_cfg: OptConfigType = None,
                 test_cfg: OptConfigType = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptConfigType = None):
        super(SELSA, self).__init__(data_preprocessor, init_cfg)
        self.detector = MODELS.build(detector)
        assert hasattr(self.detector, 'roi_head'), \
            'selsa video detector only supports two stage detector'
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg

        if frozen_modules is not None:
            self.freeze_module(frozen_modules)

    def loss(self, inputs: dict, data_samples: SampleList, **kwargs) -> dict:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size and must be 1 in SELSA method.
                The T denotes the number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'SELSA video detectors only support 1 batch size per gpu for now.'
        assert img.size(1) == 1, \
            'SELSA video detector only has 1 key image per batch.'
        img = img[0]

        ref_img = inputs['ref_img']
        assert ref_img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert ref_img.size(0) == 1, \
            'SELSA video detectors only support 1 batch size per gpu for now.'
        ref_img = ref_img[0]

        assert len(data_samples) == 1, \
            'SELSA video detectors only support 1 batch size per gpu for now.'

        all_imgs = torch.cat((img, ref_img), dim=0)
        all_x = self.detector.extract_feat(all_imgs)
        x = []
        ref_x = []
        for i in range(len(all_x)):
            x.append(all_x[i][[0]])
            ref_x.append(all_x[i][1:])

        losses = dict()
        ref_data_samples, _ = convert_data_sample_type(
            data_samples[0], num_ref_imgs=len(ref_img))

        # RPN forward and loss
        if self.detector.with_rpn:
            proposal_cfg = self.detector.train_cfg.get(
                'rpn_proposal', self.detector.test_cfg.rpn)
            rpn_data_samples = deepcopy(data_samples)
            # set cat_id of gt_labels to 0 in RPN
            for data_sample in rpn_data_samples:
                data_sample.gt_instances.labels = torch.zeros_like(
                    data_sample.gt_instances.labels)
            (rpn_losses,
             proposal_list) = self.detector.rpn_head.loss_and_predict(
                 x, rpn_data_samples, proposal_cfg=proposal_cfg)
            losses.update(rpn_losses)
            ref_proposals_list = self.detector.rpn_head.predict(
                ref_x, ref_data_samples)
        else:
            proposal_list, ref_proposals_list = [], []
            for i in range(len(data_samples)):
                proposal, ref_proposals = InstanceData(), InstanceData()
                proposal.bboxes = data_samples[i].proposals
                proposal_list.append(proposal)
                ref_proposals.bboxes = data_samples[i].ref_proposals
                ref_proposals_list.append(ref_proposals)

        roi_losses = self.detector.roi_head.loss(x, ref_x, proposal_list,
                                                 ref_proposals_list,
                                                 data_samples, **kwargs)

        losses.update(roi_losses)

        return losses

    def extract_feats(self, img: Tensor, img_metas: dict,
                      ref_img: Optional[Tensor],
                      ref_img_metas: Optional[dict]) -> Tuple:
        """Extract features for `img` during testing.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding input image.
                Typically these should be mean centered and std scaled.

            img_metas (dict): list of image information dict where each
                dict may has: 'img_id', 'img_path',
                'ori_shape', 'img_shape', 'scale_factor','flip',
                'flip_direction', 'frame_id', 'is_video_data', 'video_id',
                'video_length', 'instances'.

            ref_img (Tensor | None): of shape (1, N, C, H, W) encoding input
                reference images. Typically these should be mean centered and
                std scaled. N denotes the number of reference images. There
                may be no reference images in some cases.

            ref_img_metas (list[dict] | None): The list contains image
                information dict where each dict may has: 'img_id', 'img_path',
                'ori_shape', 'img_shape', 'scale_factor','flip',
                'flip_direction', 'frame_id', 'is_video_data', 'video_id',
                'video_length', 'instances'.

        Returns:
            tuple(x, img_metas, ref_x, ref_img_metas): x is the multi level
                feature maps of `img`, ref_x is the multi level feature maps
                of `ref_img`.
        """
        frame_id = img_metas.get('frame_id', -1)
        assert frame_id >= 0
        num_left_ref_imgs = img_metas.get('num_left_ref_imgs', -1)
        frame_stride = img_metas.get('frame_stride', -1)

        # test with adaptive stride
        if frame_stride < 1:
            if frame_id == 0:
                self.memo = Dict()
                self.memo.img_metas = ref_img_metas
                ref_x = self.detector.extract_feat(ref_img)
                # 'tuple' object (e.g. the output of FPN) does not support
                # item assignment
                self.memo.feats = []
                for i in range(len(ref_x)):
                    self.memo.feats.append(ref_x[i])

            x = self.detector.extract_feat(img)
            ref_x = self.memo.feats.copy()
            for i in range(len(x)):
                ref_x[i] = torch.cat((ref_x[i], x[i]), dim=0)
            ref_img_metas = self.memo.img_metas.copy()
            ref_img_metas.append(img_metas)
        # test with fixed stride
        else:
            if frame_id == 0:
                self.memo = Dict()
                self.memo.img_metas = ref_img_metas[0]
                ref_x = self.detector.extract_feat(ref_img)
                # 'tuple' object (e.g. the output of FPN) does not support
                # item assignment
                self.memo.feats = []
                # the features of img is same as ref_x[i][[num_left_ref_imgs]]
                x = []
                for i in range(len(ref_x)):
                    self.memo.feats.append(ref_x[i])
                    x.append(ref_x[i][[num_left_ref_imgs]])
            elif frame_id % frame_stride == 0:
                assert ref_img is not None
                x = []
                ref_x = self.detector.extract_feat(ref_img)
                for i in range(len(ref_x)):
                    self.memo.feats[i] = torch.cat(
                        (self.memo.feats[i], ref_x[i]), dim=0)[1:]
                    x.append(self.memo.feats[i][[num_left_ref_imgs]])
                self.memo.img_metas.extend(ref_img_metas[0])
                self.memo.img_metas = self.memo.img_metas[1:]
            else:
                assert ref_img is None
                x = self.detector.extract_feat(img)

            ref_x = self.memo.feats.copy()
            for i in range(len(x)):
                ref_x[i][num_left_ref_imgs] = x[i]
            ref_img_metas = self.memo.img_metas.copy()
            ref_img_metas[num_left_ref_imgs] = img_metas

        return x, img_metas, ref_x, ref_img_metas

    def predict(self,
                inputs: dict,
                data_samples: SampleList,
                rescale: bool = True) -> SampleList:
        """Test without augmentation.

        Args:
            inputs (dict[Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size and must be 1 in SELSA method.
                The T denotes the number of key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor, Optional): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            list[obj:`TrackDataSample`]: Tracking results of the
            input images. Each TrackDataSample usually contains
            ``pred_det_instances`` or ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'SELSA video detectors only support 1 batch size per gpu for now.'
        assert img.size(1) == 1, \
            'SELSA video detector only has 1 key image per batch.'
        img = img[0]

        if 'ref_img' in inputs:
            ref_img = inputs['ref_img']
            assert ref_img.dim(
            ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
            assert ref_img.size(0) == 1, 'SELSA video detectors only support' \
                                         ' 1 batch size per gpu for now.'
            ref_img = ref_img[0]
        else:
            ref_img = None

        assert len(data_samples) == 1, \
            'SELSA video detectors only support 1 batch size per gpu for now.'

        data_sample = data_samples[0]
        img_metas = data_sample.metainfo

        if ref_img is not None:
            _, ref_img_metas = convert_data_sample_type(
                data_sample, num_ref_imgs=len(ref_img))
        else:
            ref_img_metas = None

        x, img_metas, ref_x, ref_img_metas = self.extract_feats(
            img, img_metas, ref_img, ref_img_metas)

        ref_data_samples = [
            deepcopy(data_sample) for _ in range(len(ref_img_metas))
        ]
        for i in range(len(ref_img_metas)):
            ref_data_samples[i].set_metainfo(ref_img_metas[i])

        if data_samples[0].get('proposals', None) is None:
            proposal_list = self.detector.rpn_head.predict(x, data_samples)
            ref_proposals_list = self.detector.rpn_head.predict(
                ref_x, ref_data_samples)
        else:
            assert hasattr(data_samples[0], 'ref_proposals')
            proposal_list = data_samples[0].proposals
            ref_proposals_list = data_samples[0].ref_proposals

        results_list = self.detector.roi_head.predict(
            x,
            ref_x,
            proposal_list,
            ref_proposals_list,
            data_samples,
            rescale=rescale)

        track_data_sample = deepcopy(data_samples[0])
        track_data_sample.pred_det_instances = results_list[0]
        return [track_data_sample]

    def aug_test(self,
                 inputs: dict,
                 data_samples: SampleList,
                 rescale: bool = True,
                 **kwargs):
        """Test function with test time augmentation."""
        raise NotImplementedError



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, Optional

import torch
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from mmtrack.utils import OptConfigType, OptMultiConfig, SampleList
from .base import BaseMultiObjectTracker


@MODELS.register_module()
class QDTrack(BaseMultiObjectTracker):
    """Quasi-Dense Similarity Learning for Multiple Object Tracking.

    This multi object tracker is the implementation of `QDTrack
    <https://arxiv.org/abs/2006.06664>`_.

    Args:
        detector (dict): Configuration of detector. Defaults to None.
        track_head (dict): Configuration of track head. Defaults to None.
        tracker (dict): Configuration of tracker. Defaults to None.
        freeze_detector (bool): If True, freeze the detector weights.
            Defaults to False.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 detector: Optional[dict] = None,
                 track_head: Optional[dict] = None,
                 tracker: Optional[dict] = None,
                 freeze_detector: bool = False,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super().__init__(data_preprocessor, init_cfg)
        if detector is not None:
            self.detector = MODELS.build(detector)

        if track_head is not None:
            self.track_head = MODELS.build(track_head)

        if tracker is not None:
            self.tracker = MODELS.build(tracker)

        self.freeze_detector = freeze_detector
        if self.freeze_detector:
            self.freeze_module('detector')

    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> dict:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        # modify the inputs shape to fit mmdet
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(1) == 1, \
            'QDTrack can only have 1 key frame and 1 reference frame.'
        img = img[:, 0]

        ref_img = inputs['ref_img']
        assert ref_img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert ref_img.size(1) == 1, \
            'QDTrack can only have 1 key frame and 1 reference frame.'
        ref_img = ref_img[:, 0]

        x = self.detector.extract_feat(img)
        ref_x = self.detector.extract_feat(ref_img)

        losses = dict()

        # RPN forward and loss
        assert self.detector.with_rpn, \
            'QDTrack only support detector with RPN.'

        proposal_cfg = self.detector.train_cfg.get('rpn_proposal',
                                                   self.detector.test_cfg.rpn)
        rpn_data_samples = copy.deepcopy(data_samples)
        # set cat_id of gt_labels to 0 in RPN
        for data_sample in rpn_data_samples:
            data_sample.gt_instances.labels = \
                torch.zeros_like(data_sample.gt_instances.labels)

        rpn_losses, rpn_results_list = self.detector.rpn_head. \
            loss_and_predict(x,
                             rpn_data_samples,
                             proposal_cfg=proposal_cfg,
                             **kwargs)
        # avoid get same name with roi_head loss
        keys = rpn_losses.keys()
        for key in keys:
            if 'loss' in key and 'rpn' not in key:
                rpn_losses[f'rpn_{key}'] = rpn_losses.pop(key)
        losses.update(rpn_losses)

        losses_detect = self.detector.roi_head.loss(x, rpn_results_list,
                                                    data_samples, **kwargs)
        losses.update(losses_detect)

        # adjust the key of ref_img in data_samples
        ref_rpn_data_samples = []
        for data_sample in data_samples:
            ref_rpn_data_sample = TrackDataSample()
            ref_rpn_data_sample.set_metainfo(
                metainfo=dict(
                    img_shape=data_sample.metainfo['ref_img_shape'],
                    scale_factor=data_sample.metainfo['ref_scale_factor']))
            ref_rpn_data_samples.append(ref_rpn_data_sample)
        ref_rpn_results_list = self.detector.rpn_head.predict(
            ref_x, ref_rpn_data_samples, **kwargs)
        losses_track = self.track_head.loss(x, ref_x, rpn_results_list,
                                            ref_rpn_results_list, data_samples,
                                            **kwargs)
        losses.update(losses_track)

        return losses

    def predict(self,
                inputs: Dict[str, Tensor],
                data_samples: SampleList,
                rescale: bool = True,
                **kwargs) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.
            rescale (bool, Optional): If False, then returned bboxes and masks
                will fit the scale of img, otherwise, returned bboxes and masks
                will fit the scale of original image shape. Defaults to True.

        Returns:
            SampleList: Tracking results of the input images.
            Each TrackDataSample usually contains ``pred_det_instances``
            or ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(1) == 1, \
            'QDTrack can only have 1 key frame.'
        img = img[:, 0]

        assert len(data_samples) == 1, \
            'QDTrack only support 1 batch size per gpu for now.'
        metainfo = data_samples[0].metainfo
        frame_id = metainfo.get('frame_id', -1)
        if frame_id == 0:
            self.tracker.reset()

        x = self.detector.extract_feat(img)
        rpn_results_list = self.detector.rpn_head.predict(x, data_samples)
        det_results = self.detector.roi_head.predict(
            x, rpn_results_list, data_samples, rescale=rescale)

        track_data_sample = data_samples[0]
        track_data_sample.pred_det_instances = \
            det_results[0].clone()

        pred_track_instances = self.tracker.track(
            model=self,
            img=img,
            feats=x,
            data_sample=track_data_sample,
            rescale=rescale,
            **kwargs)
        track_data_sample.pred_track_instances = pred_track_instances

        return [track_data_sample]



_base_ = [
    '../../_base_/datasets/trackingnet.py',
    './stark-st1_r50_8xb16-500e_got10k-lasot-trackingnet-coco_base.py'
]

# evaluator
val_evaluator = dict(outfile_prefix='results/stark_st1_trackingnet')
test_evaluator = val_evaluator



_base_ = [
    '../../_base_/datasets/uav123.py',
    './siamese-rpn_r50_8xb28-20e_imagenetvid-imagenetdet-coco_base.py'
]

# model settings
model = dict(
    test_cfg=dict(rpn=dict(penalty_k=0.1, window_influence=0.1, lr=0.5)))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

import torch
from mmdet.models.layers import multiclass_nms
from mmdet.structures.bbox import bbox_overlaps
from mmengine.structures import InstanceData
# TODO: unify the linear_assignment package for different trackers
from scipy.optimize import linear_sum_assignment
from torch import Tensor, nn

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from mmtrack.utils import OptConfigType, imrenormalize
from .base_tracker import BaseTracker


@MODELS.register_module()
class TracktorTracker(BaseTracker):
    """Tracker for Tracktor.

    Args:
        obj_score_thr (float, optional): Threshold to filter the objects.
            Defaults to 0.5.
        reid (dict, optional): Configuration for the ReID model.

            - obj_score_thr (float, optional): Threshold to filter the
                regressed objects. Default to 0.5.
            - nms (dict, optional): NMS configuration to filter the regressed
                objects. Default to `dict(type='nms', iou_threshold=0.6)`.
            - match_iou_thr (float, optional): Minimum IoU when matching
                objects with IoU. Default to 0.3.
        reid (dict, optional): Configuration for the ReID model.

            - num_samples (int, optional): Number of samples to calculate the
                feature embeddings of a track. Default to 10.
            - image_scale (tuple, optional): Input scale of the ReID model.
                Default to (256, 128).
            - img_norm_cfg (dict, optional): Configuration to normalize the
                input. Default to None.
            - match_score_thr (float, optional): Similarity threshold for the
                matching process. Default to 2.0.
            - match_iou_thr (float, optional): Minimum IoU when matching
                objects with embedding similarity. Default to 0.2.
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Defaults to None.
    """

    def __init__(self,
                 obj_score_thr: float = 0.5,
                 regression: dict = dict(
                     obj_score_thr=0.5,
                     nms=dict(type='nms', iou_threshold=0.6),
                     match_iou_thr=0.3),
                 reid: dict = dict(
                     num_samples=10,
                     img_scale=(256, 128),
                     img_norm_cfg=None,
                     match_score_thr=2.0,
                     match_iou_thr=0.2),
                 **kwargs):
        super().__init__(**kwargs)
        self.obj_score_thr = obj_score_thr
        self.regression = regression
        self.reid = reid

    def regress_tracks(self,
                       x: List[Tensor],
                       metainfo: dict,
                       detector: nn.Module,
                       frame_id: int,
                       rescale: bool = False):
        """Regress the tracks to current frame."""
        memo = self.memo
        bboxes = memo.bboxes[memo.frame_ids == frame_id - 1]
        ids = memo.ids[memo.frame_ids == frame_id - 1]

        if rescale:
            factor_x, factor_y = metainfo['scale_factor']
            bboxes *= torch.tensor([factor_x, factor_y, factor_x,
                                    factor_y]).to(bboxes.device)

        if bboxes.size(0) == 0:
            return bboxes.new_zeros((0, 4)), bboxes.new_zeros(0), \
                   ids.new_zeros(0), ids.new_zeros(0),

        proposals = InstanceData(**dict(bboxes=bboxes))
        det_results = detector.roi_head.predict_bbox(x, [metainfo],
                                                     [proposals], None,
                                                     rescale)
        track_bboxes = det_results[0].bboxes
        track_scores = det_results[0].scores
        _track_bboxes, track_labels, valid_inds = multiclass_nms(
            track_bboxes,
            track_scores,
            0,
            self.regression['nms'],
            return_inds=True)
        ids = ids[valid_inds]

        track_bboxes = _track_bboxes[:, :-1].clone()
        track_scores = _track_bboxes[:, -1].clone()
        valid_inds = track_scores > self.regression['obj_score_thr']
        return track_bboxes[valid_inds], track_scores[
            valid_inds], track_labels[valid_inds], ids[valid_inds]

    def track(self,
              model: nn.Module,
              img: Tensor,
              feats: List[Tensor],
              data_sample: TrackDataSample,
              data_preprocessor: OptConfigType = None,
              rescale: bool = False,
              **kwargs) -> InstanceData:
        """Tracking forward function.

        Args:
            model (nn.Module): MOT model.
            img (Tensor): of shape (T, C, H, W) encoding input image.
                Typically these should be mean centered and std scaled.
                The T denotes the number of key images and usually is 1 in
                ByteTrack method.
            feats (list[Tensor]): Multi level feature maps of `img`.
            data_sample (:obj:`TrackDataSample`): The data sample.
                It includes information such as `pred_det_instances`.
            data_preprocessor (dict or ConfigDict, optional): The pre-process
               config of :class:`TrackDataPreprocessor`.  it usually includes,
                ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
            rescale (bool, optional): If True, the bounding boxes should be
                rescaled to fit the original scale of the image. Defaults to
                False.

        Returns:
            :obj:`InstanceData`: Tracking results of the input images.
            Each InstanceData usually contains ``bboxes``, ``labels``,
            ``scores`` and ``instances_id``.
        """
        metainfo = data_sample.metainfo
        bboxes = data_sample.pred_det_instances.bboxes
        labels = data_sample.pred_det_instances.labels
        scores = data_sample.pred_det_instances.scores

        frame_id = metainfo.get('frame_id', -1)
        if frame_id == 0:
            self.reset()

        if self.with_reid:
            if self.reid.get('img_norm_cfg', False):
                img_norm_cfg = dict(
                    mean=data_preprocessor['mean'],
                    std=data_preprocessor['std'],
                    to_bgr=data_preprocessor['rgb_to_bgr'])
                reid_img = imrenormalize(img, img_norm_cfg,
                                         self.reid['img_norm_cfg'])
            else:
                reid_img = img.clone()

        valid_inds = scores > self.obj_score_thr
        bboxes = bboxes[valid_inds]
        labels = labels[valid_inds]
        scores = scores[valid_inds]

        if self.empty:
            num_new_tracks = bboxes.size(0)
            ids = torch.arange(
                self.num_tracks,
                self.num_tracks + num_new_tracks,
                dtype=torch.long).to(bboxes.device)
            self.num_tracks += num_new_tracks
            if self.with_reid:
                crops = self.crop_imgs(reid_img, metainfo, bboxes.clone(),
                                       rescale)
                if crops.size(0) > 0:
                    embeds = model.reid(crops, mode='tensor')
                else:
                    embeds = crops.new_zeros((0, model.reid.head.out_channels))
        else:
            # motion
            if model.with_cmc:
                if model.with_linear_motion:
                    num_samples = model.linear_motion.num_samples
                else:
                    num_samples = 1
                self.tracks = model.cmc.track(self.last_img, img, self.tracks,
                                              num_samples, frame_id)

            if model.with_linear_motion:
                self.tracks = model.linear_motion.track(self.tracks, frame_id)

            # propagate tracks
            prop_bboxes, prop_scores, prop_labels, prop_ids = \
                self.regress_tracks(feats, metainfo,
                                    model.detector, frame_id, rescale)

            # filter bboxes with propagated tracks
            ious = bbox_overlaps(bboxes[:, :4], prop_bboxes[:, :4])
            valid_inds = (ious < self.regression['match_iou_thr']).all(dim=1)
            bboxes = bboxes[valid_inds]
            labels = labels[valid_inds]
            scores = scores[valid_inds]
            ids = torch.full((bboxes.size(0), ), -1,
                             dtype=torch.long).to(bboxes.device)

            if self.with_reid:
                prop_crops = self.crop_imgs(reid_img, metainfo,
                                            prop_bboxes.clone(), rescale)
                if prop_crops.size(0) > 0:
                    prop_embeds = model.reid(prop_crops, mode='tensor')
                else:
                    prop_embeds = prop_crops.new_zeros(
                        (0, model.reid.head.out_channels))
                if bboxes.size(0) > 0:
                    embeds = model.reid(
                        self.crop_imgs(reid_img, metainfo, bboxes.clone(),
                                       rescale),
                        mode='tensor')
                else:
                    embeds = prop_embeds.\
                        new_zeros((0, model.reid.head.out_channels))
                # reid
                active_ids = [int(_) for _ in self.ids if _ not in prop_ids]
                if len(active_ids) > 0 and bboxes.size(0) > 0:
                    track_embeds = self.get(
                        'embeds',
                        active_ids,
                        self.reid.get('num_samples', None),
                        behavior='mean')
                    reid_dists = torch.cdist(track_embeds,
                                             embeds).cpu().numpy()

                    track_bboxes = self.get('bboxes', active_ids)
                    ious = bbox_overlaps(track_bboxes, bboxes).cpu().numpy()
                    iou_masks = ious < self.reid['match_iou_thr']
                    reid_dists[iou_masks] = 1e6

                    row, col = linear_sum_assignment(reid_dists)
                    for r, c in zip(row, col):
                        dist = reid_dists[r, c]
                        if dist <= self.reid['match_score_thr']:
                            ids[c] = active_ids[r]

            new_track_inds = ids == -1
            ids[new_track_inds] = torch.arange(
                self.num_tracks,
                self.num_tracks + new_track_inds.sum(),
                dtype=torch.long).to(bboxes.device)
            self.num_tracks += new_track_inds.sum()

            bboxes = torch.cat((prop_bboxes, bboxes), dim=0)
            scores = torch.cat((prop_scores, scores), dim=0)
            labels = torch.cat((prop_labels, labels), dim=0)
            ids = torch.cat((prop_ids, ids), dim=0)
            if self.with_reid:
                embeds = torch.cat((prop_embeds, embeds), dim=0)

        self.update(
            ids=ids,
            bboxes=bboxes,
            scores=scores,
            labels=labels,
            embeds=embeds if self.with_reid else None,
            frame_ids=frame_id)
        self.last_img = img

        # update pred_track_instances
        pred_track_instances = InstanceData()
        pred_track_instances.bboxes = bboxes
        pred_track_instances.labels = labels
        pred_track_instances.scores = scores
        pred_track_instances.instances_id = ids

        return pred_track_instances



_base_ = [
    './faster-rcnn_resnet50-fpn_8xb2-4e_mot17halftrain_test-mot17halfval.py'
]
model = dict(
    backbone=dict(
        depth=101,
        init_cfg=dict(type='Pretrained',
                      checkpoint='torchvision://resnet101')),
    init_cfg=dict(
        type='Pretrained',
        checkpoint=  # noqa: E251
        'http://download.openmmlab.com/mmdetection/v2.0/faster_rcnn/faster_rcnn_r101_fpn_2x_coco/faster_rcnn_r101_fpn_2x_coco_bbox_mAP-0.398_20200504_210455-1d2dac9c.pth'  # noqa: E501
    ))
